@@ -1,6 +1,8 @@
-from pyspark.sql.dataframe import DataFrame as SparkDF
+from pyspark.sql import DataFrame as SparkDF
 from pandas import DataFrame as PandasDF
 
+from multiprocessing.pool import ThreadPool
+
 import numpy as np
 from functools import reduce
 from pyspark.sql.types import LongType, StringType
@@ -53,15 +55,30 @@ def map_spark_to_int(data):
 	return data.rdd.map(_map_function).toDF(data.columns)
 
 
-def aggregate_spark_int(data):
-	@f.udf(LongType())
-	def spark_bitwise_xor(numbers):
-		return reduce(lambda x, y: x ^ y, numbers)
+def aggregate_spark_int(data, n_jobs):
+	data.cache()
+	data_size = data.count()
+	n_jobs = min(n_jobs, round(data_size ** 0.5))
 
-	return data.agg(*[
-		spark_bitwise_xor(f.collect_list(col)).alias(col)
-		for col in data.columns
-	])
+	@f.udf(LongType())
+	def _spark_bitwise_xor(_numbers):
+		return reduce(lambda x, y: x ^ y, _numbers)
+
+	def _aggregate_spark_int(_data):
+		return _data.agg(*[
+			_spark_bitwise_xor(f.collect_list(col)).alias(col)
+			for col in _data.columns
+		])
+
+	def _union_all(_data_list):
+		return reduce(SparkDF.unionAll, _data_list)
+
+	data_list = data.randomSplit([1.0] * n_jobs)
+	pool = ThreadPool(n_jobs)
+	result_list = pool.map(_aggregate_spark_int, data_list)
+	union = _union_all(result_list)
+
+	return _aggregate_spark_int(union)
 
 
 def concat_pandas_columns(data, name='concat_column', sep='|', return_data=False):
@@ -103,14 +120,14 @@ def aggregate_pandas_int(data):
 	return data.agg(pandas_bitwise_xor, axis='rows')
 
 
-def represent_and_encode(obj):
+def represent_and_encode(obj, n_jobs):
 	class_representation = f'{obj.__class__.__name__}'
-	obj_representation = repr(make_hashable(obj))
+	obj_representation = repr(make_hashable(obj, n_jobs=n_jobs))
 	representation = f'{class_representation}|{obj_representation}'
 	return representation.encode()
 
 
-def make_hashable(obj):
+def make_hashable(obj, n_jobs):
 	if hasattr(obj, '__hashkey__'):
 		return make_hashable(obj.__hashkey__())
 	else:
@@ -124,14 +141,15 @@ def make_hashable(obj):
 			return tuple(sorted(make_hashable(e) for e in obj))
 
 		if isinstance(obj, (SparkDF, PandasDF)):
-			return make_dataframe_hashable(data=obj)
+			return make_dataframe_hashable(data=obj, n_jobs=n_jobs)
 
 		return obj
 
 
-def make_dataframe_hashable(data):
+def make_dataframe_hashable(data, n_jobs):
 	"""
 	:type data: SparkDF or PandasDF
+	:type n_jobs: int
 	:rtype: tuple
 	"""
 
@@ -156,7 +174,7 @@ def make_dataframe_hashable(data):
 
 	if use == 'spark':
 		spark_int_data = map_spark_to_int(spark_compressed)
-		spark_aggregate = aggregate_spark_int(spark_int_data)
+		spark_aggregate = aggregate_spark_int(spark_int_data, n_jobs=n_jobs)
 		aggregate_dictionary = spark_aggregate.first().asDict()
 
 	elif use == 'pandas':
